#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wss_b.h"
#include "wss_n.h"



#define NTESTS 100000

unsigned long long t1[NTESTS];
unsigned long long t2[NTESTS];
unsigned long long t3[NTESTS];
unsigned long long t4[NTESTS];
unsigned long long t5[NTESTS];
unsigned long long t6[NTESTS];



long long cpucycles(void)
{
  unsigned long long result;
  asm volatile(".byte 15;.byte 49;shlq $32,%%rdx;orq %%rdx,%%rax"
    : "=a" (result) ::  "%rdx");
  return result;
}


static unsigned long long average(unsigned long long *t1, unsigned long long *t2, size_t tlen)
{
	size_t i;
	unsigned long long acc = 0;

	for(i=0;i<tlen;i++)
		acc += t2[i] - t1[i];

	return acc/(tlen);
}


static void print_results(const char *s1, unsigned long long *t1, unsigned long long *t2, size_t tlen)
{
	printf("%s", s1);

	printf("%7llu cycles on average\n", average(t1, t2, tlen));

	return;
}

unsigned long long wss_b_test()
{
	long i;
	int j;
	unsigned char pk[WOTS_L*HASH_BYTES], sk[SEED_BYTES], masks[(WOTS_W-1)*HASH_BYTES], sig[WOTS_L*HASH_BYTES], msg[HASH_BYTES], ftn_key[HASH_BYTES];

	printf("< WSS-B speed tests >\n");

	// srand(171220);

	init_order_tables();

	for(i=0; i<NTESTS; i++)
	{
		srand(i);

		// Message generation
		for(j=0; j<HASH_BYTES; j++)
			msg[j] = rand() & 0xff;


		// Key generation 
		t1[i] = cpucycles();

		for(j=0; j<SEED_BYTES; j++)
			sk[j] = rand() & 0xff;

		for(j=0; j<HASH_BYTES; j++)
			ftn_key[j] = rand() & 0xff;

		for(j=0; j<15*HASH_BYTES; j++)
			masks[j] = rand() & 0xff;

		wss_b_pkgen(pk, sk, masks, ftn_key);

		t2[i] = cpucycles();


		// Signing
		t3[i] = cpucycles();

		wss_b_sign(sig, msg, sk, pk, masks, ftn_key);

		t4[i] = cpucycles();


		// Verification
		t5[i] = cpucycles();

		wss_b_verify(pk, sig, msg, masks, ftn_key);

		t6[i] = cpucycles(); 
	}

	print_results("- Key generation: ", t1, t2, NTESTS);
	print_results("- Signing:        ", t3, t4, NTESTS);
	print_results("- Verification:   ", t5, t6, NTESTS);
	printf("\n");

	return average(t3, t4, NTESTS);
}


//int wss_n_test()
unsigned long long wss_n_test()
{
	long i;
	int j;
	unsigned char pk[WOTS_L*HASH_BYTES], sk[SEED_BYTES], masks[(21-1)*HASH_BYTES], sig[WOTS_L*HASH_BYTES], msg[HASH_BYTES], ftn_key[HASH_BYTES];

	printf("< WSS-N speed tests >\n");

	// srand(171220);

	init_order_tables();

	for(i=0; i<NTESTS; i++)
	{
		srand(i);
		
		// Message generation
		for(j=0; j<HASH_BYTES; j++)
			msg[j] = rand() & 0xff;


		// Key generation
		t1[i] = cpucycles();

		for(j=0; j<SEED_BYTES; j++)
			sk[j] = rand() & 0xff;

		for(j=0; j<HASH_BYTES; j++)
			ftn_key[j] = rand() & 0xff;

		for(j=0; j<20*HASH_BYTES; j++)
			masks[j] = rand() & 0xff;

		wss_n_pkgen(pk, sk, masks, ftn_key);

		t2[i] = cpucycles();


		// Signing
		t3[i] = cpucycles();

		wss_n_sign(sig, msg, sk, pk, masks, ftn_key);

		t4[i] = cpucycles();


		// Verification
		t5[i] = cpucycles();

		wss_n_verify(pk, sig, msg, masks, ftn_key);

		t6[i] = cpucycles();
	}

	print_results("- Key generation: ", t1, t2, NTESTS);
	print_results("- Signing:        ", t3, t4, NTESTS);
	print_results("- Verification:   ", t5, t6, NTESTS);

	return average(t3, t4, NTESTS);
}



int main()
{
	unsigned long long tb, tn;

	tb = wss_b_test();

	tn = wss_n_test();

	printf("\n(wss-n signing time)/(wss-b signing time) = %llu(percent)\n", tn*100/tb);

	return 0;
}
