#include "params.h"
#include "prg.h"
#include "hash.h"

#include <string.h>
#include <stdlib.h>

/*
P_3             값      o_3     u_3     u'_3    v'_3
0   0   0       0       0       0       0       0
0   0   1       1       1       1       1       1
0   0   -1      -1      2       2       2       2
0   1   0       2       3       3       3       3
0   -1  0       -2      6       4       4       20
1   0   0       4       9       5       5       4
1   0   1       5       10      6       6       5
1   0   -1      3       11      7       15      6
-1  0   0       -4      18      8       8       7
-1  0   1       -3      19      9       9       8
-1  0   -1      -5      20      10      10      9
*/

unsigned char u_prime_3[256] = {0, };

unsigned char v_prime_3[256] = {0, };



/*
P_4                 값      o_4     u_4     v_4
0   0   0   0       0       0       0       0
0   0   0   1       1       1       1       1
0   0   0   -1      -1      2       2       2
0   0   1   0       2       3       3       3
0   0   -1  0       -2      6       4       4
0   1   0   0       4       9       5       5
0   1   0   1       5       10      6       6
0   1   0   -1      3       11      7       7
0   -1  0   0       -4      18      8       8
0   -1  0   1       -3      19      9       9
0   -1  0   -1      -5      20      10      20
1   0   0   0       8       27      11      11
1   0   0   1       9       28      12      12
1   0   0   -1      7       29      13      13
1   0   1   0       10      30      14      14
1   0   -1  0       6       33      15      15
-1  0   0   0       -8      54      16      16
-1  0   0   1       -7      55      17      17
-1  0   0   -1      -9      56      18      18
-1  0   1   0       -6      57      19      19
-1  0   -1  0       -10     60      20      10
*/

unsigned char u_4[256] = {0, };

unsigned char v_4[256] = {0, };



void init_order_tables()
{
	u_prime_3[1]    = 1;
	u_prime_3[255]  = 2;
	u_prime_3[2]    = 3;
	u_prime_3[254]  = 4;
	u_prime_3[4]    = 5;
	u_prime_3[5]    = 6;
	u_prime_3[3]    = 15;
	u_prime_3[252]  = 8;
	u_prime_3[253]  = 9;
	u_prime_3[251]  = 10;

	v_prime_3[1]    = 1;
	v_prime_3[255]  = 2;
	v_prime_3[2]    = 3;
	v_prime_3[254]  = 20;
	v_prime_3[4]    = 4;
	v_prime_3[5]    = 5;
	v_prime_3[3]    = 6;
	v_prime_3[252]  = 7;
	v_prime_3[253]  = 8;
	v_prime_3[251]  = 9;

	u_4[1]    = 1;
	u_4[255]  = 2;
	u_4[2]    = 3;
	u_4[254]  = 4;
	u_4[4]    = 5;
	u_4[5]    = 6;
	u_4[3]    = 7;
	u_4[252]  = 8;
	u_4[253]  = 9;
	u_4[251]  = 10;
	u_4[8]    = 11;
	u_4[9]    = 12;
	u_4[7]    = 13;
	u_4[10]   = 14;
	u_4[6]    = 15;
	u_4[248]  = 16;
	u_4[249]  = 17;
	u_4[247]  = 18;
	u_4[250]  = 19;
	u_4[246]  = 20;

	v_4[1]    = 1;
	v_4[255]  = 2;
	v_4[2]    = 3;
	v_4[254]  = 4;
	v_4[4]    = 5;
	v_4[5]    = 6;
	v_4[3]    = 7;
	v_4[252]  = 8;
	v_4[253]  = 9;
	v_4[251]  = 20;
	v_4[8]    = 11;
	v_4[9]    = 12;
	v_4[7]    = 13;
	v_4[10]   = 14;
	v_4[6]    = 15;
	v_4[248]  = 16;
	v_4[249]  = 17;
	v_4[247]  = 18;
	v_4[250]  = 19;
	v_4[246]  = 10;	
}



static void expand_seed(unsigned char outseeds[WOTS_L*HASH_BYTES], const unsigned char inseed[SEED_BYTES])
{
  prg(outseeds, WOTS_L*HASH_BYTES, inseed);
}


static void gen_chain(unsigned char out[HASH_BYTES], const unsigned char seed[HASH_BYTES], const unsigned char *masks, unsigned char ftn_key[HASH_BYTES], int chainlen)
{
  int i,j;

  if( out != seed )
    memcpy(out, seed, HASH_BYTES);

  for(i=0;i<chainlen && i<21;i++)
    hash_n_n_mask(out,out,masks+(i*HASH_BYTES), ftn_key);
}


void wss_n_pkgen(unsigned char pk[WOTS_L*HASH_BYTES], const unsigned char sk[SEED_BYTES], const unsigned char masks[(WOTS_W-1)*HASH_BYTES], unsigned char ftn_key[HASH_BYTES])
{
  int i;

  expand_seed(pk, sk);
  
  gen_chain(pk, pk, masks, ftn_key, WOTS_W-1);

  for(i=1; i<64; i++)
    gen_chain(pk+i*HASH_BYTES, pk+i*HASH_BYTES, masks, ftn_key, 20);

  gen_chain(pk+64*HASH_BYTES, pk+64*HASH_BYTES, masks, ftn_key, WOTS_W-1);
  gen_chain(pk+65*HASH_BYTES, pk+65*HASH_BYTES, masks, ftn_key, WOTS_W-1);
  gen_chain(pk+66*HASH_BYTES, pk+66*HASH_BYTES, masks, ftn_key, WOTS_W-1);
}

void wss_n_sign(unsigned char sig[WOTS_L*HASH_BYTES], const unsigned char msgg[HASH_BYTES], const unsigned char sk[SEED_BYTES], unsigned char pk[WOTS_L*HASH_BYTES], const unsigned char masks[(WOTS_W-1)*HASH_BYTES], unsigned char ftn_key[HASH_BYTES])
{
  int basew[WOTS_L], i, j, c=1320;
  unsigned char binary_array[257];
  char c_block[64] = {0,}, d_block[64] = {0,};

  unsigned char msg[32];
  memcpy(msg, msgg, 32);
  int temp =0;

  // WOT_L(67) = length of message field(64) + length of counter field(3)

#if WOTS_W == 16

  //--------------------------------------------

  for (i = 31; i >= 1; i--)
  {
    temp = (i << 1) + 1;

    switch (msg[i] & 0x3)
    {
    case 1:
      c_block[temp] += 1;
      break;

    case 3:
      c_block[temp] -= 1;

      for( j=i ; msg[j] == 255 & j>=0 ; j--)
        msg[j] = 0;

      if(j >= 0)
        msg[j]++;

      break;
    }

    switch (msg[i] & 0x6)
    {
    case 2:
      d_block[temp] += 1;
      break;

    case 6:
      d_block[temp] -= 1;

      msg[i] |= 1;

      for( j=i ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;
      
    }

    switch (msg[i] & 0xc)
    {
    case 4:
      d_block[temp] += 2;
      break;

    case 12:
      d_block[temp] -= 2;

      msg[i] |= 3;

      for( j=i ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;
    }

    switch (msg[i] & 0x18)
    {
    case 8:
      d_block[temp] += 4;
      break;

    case 24:
      d_block[temp] -= 4;

      msg[i] |= 7;

      for( j=i ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;   
    }

    c_block[temp] += ((d_block[temp]) * 2);

    temp = i << 1;

    switch (msg[i] & 0x30)
    {
    case 16:
      c_block[temp] += 1;
      break;
    case 48:
      c_block[temp] -= 1;

      msg[i] |= 15;

      for( j=i ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;
    }

    switch (msg[i] & 0x60)
    {
    case 32:
      d_block[temp] += 1;
      break;
    case 96:
      d_block[temp] -= 1;

      msg[i] |= 31;

      for( j=i ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;
    }

    switch (msg[i] & 0xc0)
    {
    case 64:
      d_block[temp] += 2;
      break;
    case 192:
      d_block[temp] -= 2;

      msg[i] = 0;

      for( j=i-1 ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;
    }

    if ((msg[i] & 0x80) == 128)
    {
      if ((msg[i - 1] & 0x1) == 0)
      {
        d_block[temp] += 4;
      }
      else
      {
        d_block[temp] -= 4;

        msg[i] = 0;

        for( j=i-1 ; msg[j] == 255 & j>=0 ; j-- )
          msg[j] = 0;

        if( j >= 0 )
          msg[j]++;
      }
    }

    c_block[temp] += ((d_block[temp]) * 2);
  }



  switch (msg[0] & 0x3)
  {
  case 1:
    c_block[1] += 1;
    break;

  case 3:
    c_block[1] -= 1;
      msg[0]++;
    break;
  }

  switch (msg[0] & 0x6)
  {
  case 2:
    d_block[1] += 1;
    break;

  case 6:
    d_block[1] -= 1;
    msg[0] &= 0xfe;
    msg[0] += 2;
    break;
  }

  switch (msg[0] & 0xc)
  {
  case 4:
    d_block[1] += 2;
    break;

  case 12:
    d_block[1] -= 2;
    msg[0] &= 0xfc;
    msg[0] += 4;
    break;
  }

  switch (msg[0] & 0x18)
  {
  case 8:
    d_block[1] += 4;
    break;

  case 24:
    d_block[1] -= 4;
    msg[0] &= 0xf8;
    msg[0] += 8;
    break;
  }

  c_block[1] += ((d_block[1]) * 2);

  switch (msg[0] & 0x30)
  {
  case 16:
    c_block[0] += 1;
    break;

  case 48:
    c_block[0] -= 1;
    msg[0] &= 0xf0;
    msg[0] += 16;
    break;
  }

  switch (msg[0] & 0x60)
  {
  case 32:
    d_block[0] += 1;
    break;

  case 96:
    d_block[0] -= 1;
    msg[0] &= 0xe0;
    msg[0] += 32;
    break;
  }

  switch (msg[0] & 0xc0)
  {
  case 64:
    d_block[0] += 2;
    break;

  case 192:
    d_block[0] -= 2;
    msg[0] = 0;
    break;
  }

  if ((msg[0] & 0x80) == 128)
    d_block[0] += 4;
    
  c_block[0] += ((d_block[0]) * 2);
  

//--------------------------------------------


  // calculate basew[0]
  if(abs(c_block[1]) < 6)
    basew[0] = u_4[(unsigned char)c_block[0]];
  else
    basew[0] = u_prime_3[(unsigned char)d_block[0]];
  c -= basew[0];

  // calculate basew[i] from i = 1 to i = 62
  for(i=1; i<63; i++)
  {
    if(abs(c_block[i+1]) < 6)
      basew[i] = v_4[(unsigned char)c_block[i]];
    else
      basew[i] = v_prime_3[(unsigned char)d_block[i]];

    c -= basew[i];
  }

  // calculate basew[63]
  basew[63] = v_4[(unsigned char)c_block[63]];
  c -= basew[63];

  // calculate counter field
  for(i=64 ;i<WOTS_L;i++)
  {
    basew[i] = c & 0xf;
    c >>= 4;
  }
  
  // generate sig[WOTS_L*HASH_BYTES] from private key(sk)
  expand_seed(sig, sk);

	if(basew[0] != 15)
	{
		gen_chain(sig, sig, masks, ftn_key, basew[i]);
	}
	else
  {
    memcpy(sig, pk, HASH_BYTES);
  }
	
  for(i=1;i<64;i++)
	{
		if(basew[i] != 20)
		{
	    gen_chain(sig+i*HASH_BYTES, sig+i*HASH_BYTES, masks, ftn_key, basew[i]);
		}
		else
    {
      memcpy(sig+(i*HASH_BYTES), pk+(i*HASH_BYTES), HASH_BYTES);
    }
	}

  for( ; i<67; i++)
	{
		if(basew[i] != 15)
		{
			gen_chain(sig+i*HASH_BYTES, sig+i*HASH_BYTES, masks, ftn_key, basew[i]);
		}
		else
    {
      memcpy(sig+(i*HASH_BYTES), pk+(i*HASH_BYTES), HASH_BYTES);
    }
	}
  
#endif
}


void wss_n_verify(unsigned char pk[WOTS_L*HASH_BYTES], const unsigned char sig[WOTS_L*HASH_BYTES], const unsigned char msgg[HASH_BYTES], const unsigned char masks[(WOTS_W-1)*HASH_BYTES], unsigned char ftn_key[HASH_BYTES])
{
  int basew[WOTS_L], i, j, c=1320;
  unsigned char binary_array[257];
  char c_block[64] = {0,}, d_block[64] = {0,};

  unsigned char msg[32];
  memcpy(msg, msgg, 32);
  int temp = 0;

#if WOTS_W == 16

  //--------------------------------------------

  for (i = 31; i >= 1; i--)
  {
    temp = (i << 1) + 1;

    switch (msg[i] & 0x3)
    {
    case 1:
      c_block[temp] += 1;
      break;
    case 3:
      c_block[temp] -= 1;

      for( j=i ; msg[j] == 255 & j>=0 ; j--)
        msg[j] = 0;

      if(j >= 0)
        msg[j]++;

      break;
    }

    switch (msg[i] & 0x6)
    {
    case 2:
      d_block[temp] += 1;
      break;
    case 6:
      d_block[temp] -= 1;

      msg[i] |= 1;

      for( j=i ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;     
    }

    switch (msg[i] & 0xc)
    {
    case 4:
      d_block[temp] += 2;
      break;
    case 12:
      d_block[temp] -= 2;

      msg[i] |= 3;

      for( j=i ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;
    }

    switch (msg[i] & 0x18)
    {
    case 8:
      d_block[temp] += 4;
      break;
    case 24:
      d_block[temp] -= 4;

      msg[i] |= 7;

      for( j=i ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++; 

      break;   
    }

    c_block[temp] += ((d_block[temp]) * 2);

    temp = i << 1;

    switch (msg[i] & 0x30)
    {
    case 16:
      c_block[temp] += 1;
      break;
    case 48:
      c_block[temp] -= 1;

      msg[i] |= 15;

      for( j=i ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;
    }

    switch (msg[i] & 0x60)
    {
    case 32:
      d_block[temp] += 1;
      break;
    case 96:
      d_block[temp] -= 1;

      msg[i] |= 31;

      for( j=i ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;
    }

    switch (msg[i] & 0xc0)
    {
    case 64:
      d_block[temp] += 2;
      break;
    case 192:
      d_block[temp] -= 2;

      msg[i] = 0;

      for( j=i-1 ; msg[j] == 255 & j>=0 ; j-- )
        msg[j] = 0;

      if( j >= 0 )
        msg[j]++;

      break;
    }

    if ((msg[i] & 0x80) == 128)
    {
      if ((msg[i - 1] & 0x1) == 0)
      {
        d_block[temp] += 4;
      }
      else
      {
        d_block[temp] -= 4;

        msg[i] = 0;

        for( j=i-1 ; msg[j] == 255 & j>=0 ; j-- )
          msg[j] = 0;

        if( j >= 0 )
          msg[j]++;
      }
    }

    c_block[temp] += ((d_block[temp]) * 2);
  }



  switch (msg[0] & 0x3)
  {
  case 1:
    c_block[1] += 1;
    break;
  case 3:
    c_block[1] -= 1;
    msg[0]++;
    break;
  }

  switch (msg[0] & 0x6)
  {
  case 2:
    d_block[1] += 1;
    break;
  case 6:
    d_block[1] -= 1;
    msg[0] &= 0xfe;
    msg[0] += 2;
    break;
  }

  switch (msg[0] & 0xc)
  {
  case 4:
    d_block[1] += 2;
    break;
  case 12:
    d_block[1] -= 2;
    msg[0] &= 0xfc;
    msg[0] += 4;
    break;
  }

  switch (msg[0] & 0x18)
  {
  case 8:
    d_block[1] += 4;
    break;
  case 24:
    d_block[1] -= 4;
    msg[0] &= 0xf8;
    msg[0] += 8;
    break;
  }

  c_block[1] += ((d_block[1]) * 2);


  switch (msg[0] & 0x30)
  {
  case 16:
    c_block[0] += 1;
    break;
  case 48:
    c_block[0] -= 1;
    msg[0] &= 0xf0;
    msg[0] += 16;
    break;
  }

  switch (msg[0] & 0x60)
  {
  case 32:
    d_block[0] += 1;
    break;
  case 96:
    d_block[0] -= 1;
    msg[0] &= 0xe0;
    msg[0] += 32;
    break;
  }

  switch (msg[0] & 0xc0)
  {
  case 64:
    d_block[0] += 2;
    break;
  case 192:
    d_block[0] -= 2;
    msg[0] = 0;
    break;
  }

  if ((msg[0] & 0x80) == 128)
    d_block[0] += 4;
    
  c_block[0] += ((d_block[0]) * 2);
  

//--------------------------------------------

  // calculate basew[0]
  if(abs(c_block[1]) < 6)
    basew[0] = u_4[(unsigned char)c_block[0]];
  else
    basew[0] = u_prime_3[(unsigned char)d_block[0]];
  
  c -= basew[0];

  // calculate basew[i] from i = 1 to i = 62
  for(i=1; i<63; i++)
  {
    if(abs(c_block[i+1]) < 6)
      basew[i] = v_4[(unsigned char)c_block[i]];
    else
      basew[i] = v_prime_3[(unsigned char)d_block[i]];

    c -= basew[i];
  }

  // calculate basew[63]
  basew[63] = v_4[(unsigned char)c_block[63]];
  c -= basew[63];

  // calculate counter field
  for(i=64 ;i<WOTS_L;i++)
  {
    basew[i] = c & 0xf;
    c >>= 4;
  }

  gen_chain(pk, sig, masks+(basew[0]*HASH_BYTES), ftn_key, WOTS_W-1-basew[0]);

  for(i=1;i<64;i++)
	{
    gen_chain(pk+i*HASH_BYTES, sig+i*HASH_BYTES, masks+(basew[i]*HASH_BYTES), ftn_key, 20-basew[i]);
	}

  for( ;i<67;i++)
	{
    gen_chain(pk+i*HASH_BYTES, sig+i*HASH_BYTES, masks+(basew[i]*HASH_BYTES), ftn_key, WOTS_W-1-basew[i]);
	}

#endif

}
